from typing import Dict, Set
from hub.util.tag import process_hub_path
from hub.util.hash import hash_inputs
from hub.constants import WANDB_JSON_FILENMAE
from hub.hooks import (
    add_create_dataset_hook,
    add_load_dataset_hook,
    add_write_dataset_hook,
    add_read_dataset_hook,
    add_commit_dataset_hook,
)
import importlib
import sys
import json
import warnings
from hub.client.log import logger

_WANDB_INSTALLED = bool(importlib.util.find_spec("wandb"))


def wandb_run():
    return getattr(sys.modules.get("wandb"), "run", None)


_READ_DATASETS: Dict[str, Set[str]] = {}
_WRITTEN_DATASETS: Dict[str, Set[str]] = {}


def dataset_created(ds):
    pass


def dataset_loaded(ds):
    pass


def artifact_name_from_ds_path(ds) -> str:
    path = ds.path
    hash = hash_inputs(path)
    if path.startswith("hub://"):
        _, org, ds_name, _ = process_hub_path(path)
        artifact_name = f"hub-{org}-{ds_name}"
        if "/.queries/" in path:
            vid = path.split("/.queries/", 1)[1]
            artifact_name += f"-view-{vid}"
    else:
        pfix = path.split("://", 1)[0] if "://" in path else "local"
        artifact_name = f"{pfix}"
    artifact_name += f"-commit-{ds.commit_id}"
    artifact_name += f"-{hash[:8]}"
    return artifact_name


def artifact_from_ds(ds):
    import wandb

    path = ds.path
    name = artifact_name_from_ds_path(ds)
    artifact = wandb.Artifact(name, "dataset")
    artifact.add_reference(path, name="url")
    return artifact


def get_ds_key(ds):
    entry = getattr(ds, "_view_entry", None)
    if entry:
        return hash_inputs(entry)
    return (hash_inputs(ds.path, ds.commit_id),)


def dataset_config(ds):
    if hasattr(ds, "_view_entry"):
        entry = ds._view_entry
        source_ds_path = entry.source_dataset_path
        commit_id = entry.info["source-dataset-version"]
        vid = entry.id
        ret = {
            "Dataset": source_ds_path,
            "Commit ID": commit_id,
            "View ID": vid,
        }
        if source_ds_path.startswith("hub://") and ds.path.startswith("hub://"):
            ret["URL"] = _plat_url(ds)
        q = entry.query
        if q:
            ret["Query"] = q
        if entry.virtual:
            ret["Index"] = ds.index.to_json()
        else:
            ret["Index"] = list(ds.sample_indices)
        return ret

    ret = {
        "Dataset": ds.path,
        "Commit ID": ds.commit_id,
    }
    if ds.path.startswith("hub://"):
        ret["URL"] = _plat_url(ds)
    if not ds.index.is_trivial():
        ret["Index"] = ds.index.to_json()
    q = getattr(ds, "_query", None)
    if q:
        ret["Query"] = q
    return ret


def log_dataset(dsconfig):
    url = dsconfig.get("URL")
    if not url:
        return
    import wandb

    run = wandb.run
    url_prefix = "https://app.activeloop.ai/"
    url = url[len(url_prefix) :]
    # TODO : commit and view id are not supported by visualizer. Remove below line once they are supported.
    url = "/".join(url.split("/")[:2])
    run.log({f"Hub Dataset - {url}": wandb.Html(_viz_html("hub://" + url))}, step=0)


def dataset_written(ds):
    pass


def dataset_committed(ds):
    run = wandb_run()
    key = get_ds_key(ds)
    if run:
        if run.id not in _WRITTEN_DATASETS:
            _WRITTEN_DATASETS.clear()
            _WRITTEN_DATASETS[run.id] = {}
        keys = _WRITTEN_DATASETS[run.id]
        if key not in keys:
            keys[key] = None
            try:
                output_datasets = run.config.output_datasets
            except (KeyError, AttributeError):
                output_datasets = []
            dsconfig = dataset_config(ds)
            output_datasets.append(dsconfig)
            log_dataset(dsconfig)
            run.config.output_datasets = output_datasets
            artifact = artifact_from_ds(ds)
            wandb_info = read_json(ds)
            try:
                commits = wandb_info["commits"]
            except KeyError:
                commits = {}
                wandb_info["commits"] = commits
            info = {}
            commits[ds.commit_id] = info
            info["created-by"] = {
                "run": {
                    "entity": run.entity,
                    "project": run.project,
                    "id": run.id,
                    "url": run.url,
                },
                "artifact": artifact.name,
            }
            write_json(ds, wandb_info)
            run.log_artifact(artifact)


def _filter_input_datasets(input_datasets):
    ret = []
    for i, dsconfig in enumerate(input_datasets):
        if "Index" not in dsconfig:
            rm = False
            for j, dsconfig2 in enumerate(input_datasets):
                if (
                    i != j
                    and dsconfig2["Dataset"] == dsconfig["Dataset"]
                    and dsconfig2["Commit ID"] == dsconfig["Commit ID"]
                ):
                    rm = True
                    break
            if not rm:
                ret.append(dsconfig)
        else:
            ret.append(dsconfig)
    return ret


def dataset_read(ds):
    run = wandb_run()
    if not run:
        return
    if run.id not in _READ_DATASETS:
        _READ_DATASETS.clear()
        _READ_DATASETS[run.id] = {}
    keys = _READ_DATASETS[run.id]
    key = get_ds_key(ds)
    idx = ds.index.to_json()
    if key not in keys or idx not in keys[key]:
        if key not in keys:
            keys[key] = []
        keys[key].append(idx)
        try:
            input_datasets = run.config.input_datasets
        except (KeyError, AttributeError):
            input_datasets = []
        dsconfig = dataset_config(ds)
        if dsconfig not in input_datasets:
            input_datasets.append(dsconfig)
            input_datasets = _filter_input_datasets(input_datasets)
            for config in input_datasets:
                log_dataset(config)
            run.config.input_datasets = input_datasets
        if run._settings.mode != "online":
            return
        if hasattr(ds, "_view_entry") and not ds._view_entry._external:
            # TODO handle external otimized views
            ds = ds._view_entry._ds
        wandb_info = read_json(ds).get("commits", {}).get(ds.commit_id)
        if wandb_info:
            try:
                run_and_artifact = wandb_info["created-by"]
                run_info = run_and_artifact["run"]
                artifact = run_and_artifact["artifact"]
                artifact_path = (
                    f"{run_info['entity']}/{run_info['project']}/{artifact}:latest"
                )
                run.use_artifact(artifact_path)
            except Exception as e:
                warnings.warn(
                    f"Wandb integration: Error while using wandb artifact: {e}"
                )
        else:
            # For datasets that were not created during a wandb run,
            # we want to "use" an artifact that is not logged by any run.
            # This is not possible with wandb yet.

            # artifact = artifact_from_ds(ds)
            # run.use_artifact(artifact)
            pass


def _viz_html(hub_path: str):
    #     return f"""
    #       <div id='container'></div>
    #   <script src="https://app.activeloop.ai/visualizer/vis.js"></script>
    #   <script>
    #     let container = document.getElementById('container')

    #     window.vis.visualize('{hub_path}', null, null, container, {{
    #       requireSignin: true
    #     }})
    #   </script>
    #     """
    return f"""<iframe width="100%" height="100%" sandbox="allow-same-origin allow-scripts allow-popups allow-forms" src="https://app.activeloop.ai/visualizer/iframe?url={hub_path}" />"""


def _plat_url(ds, http=True):
    prefix = "https://app.activeloop.ai/" if http else "hub://"
    if hasattr(ds, "_view_entry"):
        entry = ds._view_entry
        _, org, ds_name, _ = process_hub_path(entry.source_dataset_path)
        commit_id = entry.info["source-dataset-version"]
        return f"{prefix}{org}/{ds_name}/{commit_id}?view={entry.id}"
    _, org, ds_name, _ = process_hub_path(ds.path)
    ret = f"{prefix}{org}/{ds_name}"
    if ds.commit_id:
        ret += f"/{ds.commit_id}"
    return ret


def link_html(hub_path):
    return f"""<a href="{_plat_url(hub_path)}">{hub_path}</a>"""


if _WANDB_INSTALLED:
    add_create_dataset_hook(dataset_created, "wandb_dataset_create")
    add_load_dataset_hook(dataset_loaded, "wandb_dataset_load")
    add_write_dataset_hook(dataset_written, "wandb_dataset_write")
    add_read_dataset_hook(dataset_read, "wandb_dataset_read")
    add_commit_dataset_hook(dataset_committed, "wandb_dataset_commit")


def read_json(ds):
    try:
        return json.loads(ds.base_storage[WANDB_JSON_FILENMAE].decode("utf-8"))
    except KeyError:
        return {}


def write_json(ds, dat):
    ds.base_storage[WANDB_JSON_FILENMAE] = json.dumps(dat).encode("utf-8")



from hub.experimental.convert_to_hub3 import dataset_to_hub3
from hub.util.bugout_reporter import hub_reporter


@hub_reporter.record_call
def query(dataset, query_string: str):
    """
    Query the dataset.
    """
    ds = dataset_to_hub3(dataset)
    dsv = ds.query(query_string)
    indexes = dsv.indexes  # TODO: enable this in indra
    return dataset[indexes]



from typing import Callable, List, Optional
from hub.experimental.convert_to_hub3 import dataset_to_hub3  # type: ignore
from hub.experimental.util import raise_indra_installation_error  # type: ignore
from hub.experimental.util import collate_fn as default_collate  # type: ignore
from hub.experimental.hub3_query import query
from hub.util.bugout_reporter import hub_reporter

try:
    from indra import Loader  # type: ignore

    INDRA_INSTALLED = True
except ImportError:
    INDRA_INSTALLED = False


class Hub3DataLoader:
    def __init__(
        self,
        dataset,
        _batch_size=None,
        _shuffle=None,
        _num_threads=None,
        _num_workers=None,
        _collate=None,
        _transform=None,
        _distributed=None,
        _prefetch_factor=None,
        _tensors=None,
        _drop_last=False,
        _mode=None,
    ):
        raise_indra_installation_error(INDRA_INSTALLED)
        # verifies underlying storage
        dataset_to_hub3(dataset)
        self.dataset = dataset
        self._batch_size = _batch_size
        self._shuffle = _shuffle
        self._num_threads = _num_threads
        self._num_workers = _num_workers
        self._collate = _collate
        self._transform = _transform
        self._distributed = _distributed
        self._prefetch_factor = _prefetch_factor
        self._tensors = _tensors
        self._drop_last = _drop_last
        self._mode = _mode

    def batch(self, batch_size: int, drop_last: bool = False):
        """Returns a batched DataLoader object.

        Args:
            batch_size (int): Number of samples in each batch.
            drop_last (bool): If True, the last batch will be dropped if its size is less than batch_size. Defaults to False.

        Returns:
            Dataloader: A Dataloader object.

        Raises:
            ValueError: If .batch() has already been called.
        """
        if self._batch_size is not None:
            raise ValueError("batch size is already set")

        all_vars = self.__dict__.copy()
        all_vars["_batch_size"] = batch_size
        all_vars["_drop_last"] = drop_last
        return self.__class__(**all_vars)

    def shuffle(self):
        """Returns a shuffled Dataloader object.

        Returns:
            Dataloader: A Dataloader object.

        Raises:
            ValueError: If .shuffle() has already been called.
        """
        if self._shuffle is not None:
            raise ValueError("shuffle is already set")
        all_vars = self.__dict__.copy()
        all_vars["_shuffle"] = True
        return self.__class__(**all_vars)

    def transform(self, transform_fn: Callable):
        """Returns a transformed Dataloader object.

        Args:
            transform_fn (Callable): A function that takes a sample as input and returns a transformed sample.

        Returns:
            Dataloader: A Dataloader object.

        Raises:
            ValueError: If .transform() has already been called.
        """
        if self._transform is not None:
            raise ValueError("transform is already set")
        all_vars = self.__dict__.copy()
        all_vars["_transform"] = transform_fn
        return self.__class__(**all_vars)

    def query(self, query_string: str):
        all_vars = self.__dict__.copy()
        all_vars["dataset"] = query(self.dataset, query_string)
        return self.__class__(**all_vars)

    def pytorch(
        self,
        num_workers: int = 0,
        collate_fn: Optional[Callable] = None,
        tensors: Optional[List[str]] = None,
        num_threads: Optional[int] = None,
        prefetch_factor: int = 10,
        distributed: bool = False,
    ):
        """Returns a pytorch Dataloader object.

        Args:
            num_workers (int): Number of workers to use for transforming and processing the data. Defaults to 0.
            collate_fn (Callable, Optional): merges a list of samples to form a mini-batch of Tensor(s).
            tensors (List[str], Optional): List of tensors to load. If None, all tensors are loaded. Defaults to None.
            num_threads (int, Optional): Number of threads to use for fetching and decompressing the data. If None, the number of threads is automatically determined. Defaults to None.
            prefetch_factor (int): Number of batches to transform and collate in advance per worker. Defaults to 10.
            distributed (bool): Used for DDP training. Distributes different sections of the dataset to different ranks. Defaults to False.

        Returns:
            Dataloader: A Dataloader object.

        Raises:
            ValueError: If .to_pytorch() or .to_numpy() has already been called.
        """
        if self._mode is not None:
            if self._mode == "numpy":
                raise ValueError("Can't call .to_pytorch after .to_numpy()")
            raise ValueError("already called .to_pytorch()")
        all_vars = self.__dict__.copy()
        all_vars["_num_workers"] = num_workers
        all_vars["_collate"] = collate_fn
        all_vars["_tensors"] = tensors
        all_vars["_num_threads"] = num_threads
        all_vars["_prefetch_factor"] = prefetch_factor
        all_vars["_distributed"] = distributed
        all_vars["_mode"] = "pytorch"
        return self.__class__(**all_vars)

    def numpy(
        self,
        num_workers: int = 0,
        tensors: Optional[List[str]] = None,
        num_threads: Optional[int] = None,
        prefetch_factor: int = 10,
    ):
        """Returns a numpy Dataloader object.

        Args:
            num_workers (int): Number of workers to use for transforming and processing the data. Defaults to 0.
            tensors (List[str], Optional): List of tensors to load. If None, all tensors are loaded. Defaults to None.
            num_threads (int, Optional): Number of threads to use for fetching and decompressing the data. If None, the number of threads is automatically determined. Defaults to None.
            prefetch_factor (int): Number of batches to transform and collate in advance per worker. Defaults to 10.

        Returns:
            Dataloader: A Dataloader object.

        Raises:
            ValueError: If .to_pytorch() or .to_numpy() has already been called.
        """
        if self._mode is not None:
            if self._mode == "pytorch":
                raise ValueError("Can't call .to_numpy after .to_pytorch()")
            raise ValueError("already called .to_numpy()")
        all_vars = self.__dict__.copy()
        all_vars["_num_workers"] = num_workers
        all_vars["_tensors"] = tensors
        all_vars["_num_threads"] = num_threads
        all_vars["_prefetch_factor"] = prefetch_factor
        all_vars["_mode"] = "numpy"
        return self.__class__(**all_vars)

    def __iter__(self):
        dataset = dataset_to_hub3(self.dataset)
        batch_size = self._batch_size or 1
        drop_last = self._drop_last or False

        shuffle = self._shuffle or False

        transform_fn = self._transform

        num_workers = self._num_workers or 0
        if self._collate is None and self._mode == "pytorch":
            collate_fn = default_collate
        else:
            collate_fn = self._collate
        tensors = self._tensors or []
        num_threads = self._num_threads
        prefetch_factor = self._prefetch_factor
        distributed = self._distributed or False
        upcast = (
            self._mode == "pytorch"
        )  # only upcast for pytorch, this handles unsupported dtypes
        return iter(
            Loader(
                dataset,
                batch_size=batch_size,
                num_threads=num_threads,
                shuffle=shuffle,
                num_workers=num_workers,
                collate_fn=collate_fn,
                transform_fn=transform_fn,
                distributed=distributed,
                prefetch_factor=prefetch_factor,
                tensors=tensors,
                drop_last=drop_last,
                upcast=upcast,
            )
        )


@hub_reporter.record_call
def dataloader(dataset) -> Hub3DataLoader:
    return Hub3DataLoader(dataset)
